###############################################################################
##
##  Copyright (C) 2013 Tavendo GmbH
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##      http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.
##
###############################################################################


import zope
from zope.interface import Interface, Attribute


class IWebSocketChannel(Interface):
   """
   A WebSocket channel is a bidirectional, full-duplex, ordered, reliable message channel
   over a WebSocket connection as specified in RFC6455.

   This interface defines a message-based API to WebSocket plus auxiliary hooks
   and methods.
   """

   def onConnect(requestOrResponse):
      """
      Callback fired during WebSocket opening handshake when a client connects (with
      request from client) or when server connection established (with response from
      server).

      :param requestOrResponse: Connection request or response.
      :type requestOrResponse: Instance of :class:`autobahn.websocket.protocol.ConnectionRequest`
                               or :class:`autobahn.websocket.protocol.ConnectionResponse`.
      """

   def onOpen():
      """
      Callback fired when the initial WebSocket opening handshake was completed.
      You now can send and receive WebSocket messages.
      """

   def sendMessage(payload, isBinary = False, fragmentSize = None, sync = False, doNotCompress = False):
      """
      Send a WebSocket message.

      You can send text or binary messages, and optionally specifiy a payload fragment size.
      When the latter is given, the payload will be split up into WebSocket frames each with
      payload length `<= fragmentSize`.

      :param payload: The message payload.
      :type payload: bytes
      :param isBinary: `True` iff payload is binary, else the payload must be UTF-8 encoded text.
      :type isBinary: bool
      :param fragmentSize: Fragment message into WebSocket fragments of this size.
      :type fragmentSize: int
      :param sync: Iff `True`, try to force data onto the wire immediately. Note: do NOT use
                   this normally unless you know what you are doing. Performance likely will
                   suffer significantly. This feature is mainly here for use by Autobahn|Testsuite.
      :type sync: bool
      :param doNotCompress: Iff `True`, never compress this message. This only applies to
                            Hybi-Mode and only when WebSocket compression has been negotiated on
                            the WebSocket connection. Use when you know the payload 
                            uncompressible (e.g. encrypted or already compressed).
      :type doNotCompress: bool
      """

   def onMessage(payload, isBinary):
      """
      Callback fired when a complete WebSocket message was received.

      :param payload: Message payload (UTF-8 encoded text or binary). Can also be empty when
                      the WebSocket message contained no payload.
      :type payload: bytes
      :param isBinary: `True` iff payload is binary, else the payload is UTF-8 encoded text.
      :type isBinary: bool
      """

   def sendClose(code = None, reason = None):
      """
      Starts a WebSocket closing handshake tearing down the WebSocket connection.

      :param code: An optional close status code (`1000` for normal close or `3000-4999` for
                   application specific close).
      :type code: int
      :param reason: An optional close reason (a string that when present, a status
                     code MUST also be present).
      :type reason: str
      """

   def onClose(wasClean, code, reason):
      """
      Callback fired when the WebSocket connection has been closed (WebSocket closing
      handshake has been finished or the connection was closed uncleanly).

      :param wasClean: True, iff the WebSocket connection was closed cleanly.
      :type wasClean: bool
      :param code: None or close status code (as sent by the WebSocket peer).
      :type code: int
      :param reason: None or close reason (as sent by the WebSocket peer).
      :type reason: str
      """

   def sendPreparedMessage(preparedMsg):
      """
      Send a message that was previously prepared with :func:`autobahn.websocket.protocol.WebSocketFactory.prepareMessage`.

      :param prepareMessage: A previsouly prepared message.
      :type prepareMessage: Instance of :class:`autobahn.websocket.protocol.PreparedMessage`.
      """

   def sendPing(payload = None):
      """
      Send a WebSocket ping to the peer.

      A peer is expected to pong back the payload a soon as "practical". When more than
      one ping is outstanding at a peer, the peer may elect to respond only to the last ping.

      :param payload: An (optional) arbitrary payload of length `<126` octets.
      :type payload: bytes
      """

   def onPing(payload):
      """
      Callback fired when a WebSocket ping was received. A default implementation responds
      by sending a WebSocket pong.

      :param payload: Payload of ping (when there was any). Can be arbitrary, up to `125` octets.
      :type payload: bytes
      """

   def sendPong(payload = None):
      """
      Send a WebSocket pong to the peer.

      A WebSocket pong may be sent unsolicited. This serves as a unidirectional heartbeat.
      A response to an unsolicited pong is "not expected".

      :param payload: An (optional) arbitrary payload of length < 126 octets.
      :type payload: bytes
      """

   def onPong(payload):
      """
      Callback fired when a WebSocket pong was received. A default implementation does nothing.

      :param payload: Payload of pong (when there was any). Can be arbitrary, up to 125 octets.
      :type payload: bytes
      """


class IWebSocketChannelFrameApi(IWebSocketChannel):
   """
   Frame-based API to a WebSocket channel.
   """

   def onMessageBegin(isBinary):
      """
      Callback fired when receiving of a new WebSocket message has begun.

      :param isBinary: `True` iff payload is binary, else the payload is UTF-8 encoded text.
      :type isBinary: bool
      """

   def onMessageFrame(payload):
      """
      Callback fired when a complete WebSocket message frame for a previously begun
      WebSocket message has been received.

      :param payload: Message frame payload (a list of chunks received).
      :type payload: list of bytes
      """

   def onMessageEnd():
      """
      Callback fired when a WebSocket message has been completely received (the last
      WebSocket frame for that message has been received).
      """

   def beginMessage(isBinary = False, doNotCompress = False):
      """
      Begin sending a new WebSocket message.

      :param isBinary: `True` iff payload is binary, else the payload must be UTF-8 encoded text.
      :type isBinary: bool
      :param doNotCompress: Iff `True`, never compress this message. This only applies to
                            Hybi-Mode and only when WebSocket compression has been negotiated on
                            the WebSocket connection. Use when you know the payload 
                            uncompressible (e.g. encrypted or already compressed).
      :type doNotCompress: bool
      """

   def sendMessageFrame(payload, sync = False):
      """
      When a message has been previously begun, send a complete message frame in one go.

      :param payload: The message frame payload. When sending a text message, the payload must
                      be UTF-8 encoded already.
      :type payload: bytes
      :param sync: Iff `True`, try to force data onto the wire immediately. Note: do NOT use
                   this normally unless you know what you are doing. Performance likely will
                   suffer significantly. This feature is mainly here for use by Autobahn|Testsuite.
      :type sync: bool
      """

   def endMessage():
      """
      End a message previously begun message. No more frames may be sent (for that message).
      You have to begin a new message before sending again.
      """



class IWebSocketChannelStreamingApi(IWebSocketChannelFrameApi):
   """
   Streaming API to a WebSocket channel.
   """

   def onMessageFrameBegin(length):
      """
      Callback fired when receiving a new message frame has begun.
      A default implementation will prepare to buffer message frame data.

      :param length: Payload length of message frame which is subsequently received.
      :type length: int
      """

   def onMessageFrameData(payload):
      """
      Callback fired when receiving data within a previously begun message frame.
      A default implementation will buffer data for frame.

      :param payload: Partial payload for message frame.
      :type payload: bytes
      """

   def onMessageFrameEnd():
      """
      Callback fired when a previously begun message frame has been completely received.
      A default implementation will flatten the buffered frame data and
      fire `onMessageFrame`.
      """

   def beginMessageFrame(length):
      """
      Begin sending a new message frame.

      :param length: Length of the frame which is to be started. Must be `>= 0` and `<= 2^63`.
      :type length: int
      """

   def sendMessageFrameData(payload, sync = False):
      """
      Send out data when within a message frame (message was begun, frame was begun).
      Note that the frame is automatically ended when enough data has been sent.
      In other words, there is no `endMessageFrame`, since you have begun the frame
      specifying the frame length, which implicitly defined the frame end. This is different
      from messages, which you begin *and* end explicitly , since a message can contain
      an unlimited number of frames.

      :param payload: Frame payload to send.
      :type payload: bytes
      :param sync: Iff `True`, try to force data onto the wire immediately. Note: do NOT use
                   this normally unless you know what you are doing. Performance likely will
                   suffer significantly. This feature is mainly here for use by Autobahn|Testsuite.
      :type sync: bool

      :returns: int -- When the currently sent message frame is still incomplete,
                       returns octets remaining to be sent. When the frame is complete,
                       returns `0`, when `< 0`, the amount of unconsumed data in payload
                       argument.
      """



class IWampDealer(Interface):
   """
   """

   def register(self, endpoint, obj):
      """
      """

   def registerMethod(self, endpoint, obj, method):
      """
      """

   def registerProcedure(self, endpoint, procedure):
      """
      """

   def unregister(self, endpoint):
      """
      """


class IWampBroker(Interface):
   """
   """

   def register(self, topic, prefix = False, publish = True, subscribe = True):
      """
      """

   def unregister(self, topic):
      """
      """


class IWampPublishOptions(Interface):

   excludeMe = Attribute("Exclude me, the publisher, from receiving the event (even though I may be subscribed).")



class IWampSession(Interface):
   """
   """

   def call(self, *args):
      """
      """

   def subscribe(self, topic, handler):
      """
      """

   def unsubscribe(self, topic, handler = None):
      """
      """

   def publish(self, topic, event,
               excludeMe = None,
               exclude = None,
               eligible = None,
               discloseMe = None):
      """
      """

   def setDealer(self, dealer = None):
      """
      """

   def setBroker(self, broker = None):
      """
      """



###############################################################################
##
##  Copyright 2013 Tavendo GmbH
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##      http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.
##
###############################################################################

__all__ = ("FlashPolicyProtocol", "FlashPolicyFactory",)


import re

from twisted.internet.protocol import Protocol, Factory


class FlashPolicyProtocol(Protocol):
   """
   Flash Player 9 (version 9.0.124.0 and above) implements a strict new access
   policy for Flash applications that make Socket or XMLSocket connections to
   a remote host. It now requires the presence of a socket policy file
   on the server.

   We want this to support the Flash WebSockets bridge which is needed for
   older browser, in particular MSIE9/8.

   See:
      * `Autobahn WebSocket fallbacks example <https://github.com/tavendo/AutobahnPython/tree/master/examples/twisted/websocket/echo_wsfallbacks>`_
      * `Flash policy files background <http://www.lightsphere.com/dev/articles/flash_socket_policy.html>`_
   """

   REQUESTPAT = re.compile("^\s*<policy-file-request\s*/>")
   REQUESTMAXLEN = 200
   REQUESTTIMEOUT = 5
   POLICYFILE = """<?xml version="1.0"?><cross-domain-policy><allow-access-from domain="*" to-ports="%d" /></cross-domain-policy>"""

   def __init__(self, allowedPort):
      """
      Ctor.

      :param allowedPort: The port to which Flash player should be allowed to connect.
      :type allowedPort: int
      """
      self.allowedPort = allowedPort
      self.received = ""
      self.dropConnection = None


   def connectionMade(self):
      ## DoS protection
      ##
      def dropConnection():
         self.transport.abortConnection()
         self.dropConnection = None
      self.dropConnection = self.factory.reactor.callLater(FlashPolicyProtocol.REQUESTTIMEOUT, dropConnection)


   def connectionLost(self, reason):
      if self.dropConnection:
         self.dropConnection.cancel()
         self.dropConnection = None


   def dataReceived(self, data):
      self.received += data
      if FlashPolicyProtocol.REQUESTPAT.match(self.received):
         ## got valid request: send policy file
         ##
         self.transport.write(FlashPolicyProtocol.POLICYFILE % self.allowedPort)
         self.transport.loseConnection()
      elif len(self.received) > FlashPolicyProtocol.REQUESTMAXLEN:
         ## possible DoS attack
         ##
         self.transport.abortConnection()
      else:
         ## need more data
         ##
         pass


class FlashPolicyFactory(Factory):

   def __init__(self, allowedPort, reactor = None):
      """
      Ctor.

      :param allowedPort: The port to which Flash player should be allowed to connect.
      :type allowedPort: int
      :param reactor: Twisted reactor to use. If not given, autoimport.
      :type reactor: obj
      """
      ## lazy import to avoid reactor install upon module import
      if reactor is None:
         from twisted.internet import reactor
      self.reactor = reactor

      self.allowedPort = allowedPort


   def buildProtocol(self, addr):
      proto = FlashPolicyProtocol(self.allowedPort)
      proto.factory = self
      return proto



###############################################################################
##
##  Copyright (C) 2013 Tavendo GmbH
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##      http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.
##
###############################################################################

__all__ = ['WebSocketServerProtocol',
           'WebSocketServerFactory',
           'WebSocketClientProtocol',
           'WebSocketClientFactory']

from collections import deque

import asyncio
from asyncio.tasks import iscoroutine
from asyncio import Future

from autobahn.websocket import protocol
from autobahn.websocket import http


def yields(value):
   """
   Return True iff the value yields.

   See: http://stackoverflow.com/questions/20730248/maybedeferred-analog-with-asyncio
   """
   return isinstance(value, Future) or iscoroutine(value)



class WebSocketAdapterProtocol(asyncio.Protocol):
   """
   Adapter class for Asyncio WebSocket client and server protocols.
   """

   def connection_made(self, transport):
      self.transport = transport

      self.receive_queue = deque()
      self.waiter = Future()
      asyncio.Task(self._consume())

      peer = transport.get_extra_info('peername')
      try:
         self.peer = "%s:%d" % (peer[0], peer[1])
      except:
         ## eg Unix Domain sockets don't have host/port
         self.peer = str(peer)

      self._connectionMade()


   def connection_lost(self, exc):
      self._connectionLost(exc)
      self.transport = None


   def _consume(self):
      while True:
         yield from self.waiter
         self.waiter = Future()
         while len(self.receive_queue):
            data = self.receive_queue.popleft()
            if self.transport:
               try:
                  self._dataReceived(data)
               except Exception as e:
                  raise e
                  #print("WebSocketAdapterProtocol._consume: {}".format(e))
            else:
               print("WebSocketAdapterProtocol._consume: no transport")


   def data_received(self, data):
      self.receive_queue.append(data)
      if not self.waiter.done():
         self.waiter.set_result(None)


   def _closeConnection(self, abort = False):
      self.transport.close()


   def _onOpen(self):
      res = self.onOpen()
      if yields(res):
         asyncio.async(res)

   def _onMessageBegin(self, isBinary):
      res = self.onMessageBegin(isBinary)
      if yields(res):
         asyncio.async(res)

   def _onMessageFrameBegin(self, length):
      res = self.onMessageFrameBegin(length)
      if yields(res):
         asyncio.async(res)

   def _onMessageFrameData(self, payload):
      res = self.onMessageFrameData(payload)
      if yields(res):
         asyncio.async(res)

   def _onMessageFrameEnd(self):
      res = self.onMessageFrameEnd()
      if yields(res):
         asyncio.async(res)

   def _onMessageFrame(self, payload):
      res = self.onMessageFrame(payload)
      if yields(res):
         asyncio.async(res)

   def _onMessageEnd(self):
      res = self.onMessageEnd()
      if yields(res):
         asyncio.async(res)

   def _onMessage(self, payload, isBinary):
      res = self.onMessage(payload, isBinary)
      if yields(res):
         asyncio.async(res)

   def _onPing(self, payload):
      res = self.onPing(payload)
      if yields(res):
         asyncio.async(res)

   def _onPong(self, payload):
      res = self.onPong(payload)
      if yields(res):
         asyncio.async(res)

   def _onClose(self, wasClean, code, reason):
      res = self.onClose(wasClean, code, reason)
      if yields(res):
         asyncio.async(res)


   def registerProducer(self, producer, streaming):
      raise Exception("not implemented")



class WebSocketServerProtocol(WebSocketAdapterProtocol, protocol.WebSocketServerProtocol):
   """
   Base class for Asyncio WebSocket server protocols.
   """

   def _onConnect(self, request):
      ## onConnect() will return the selected subprotocol or None
      ## or a pair (protocol, headers) or raise an HttpException
      ##
      try:
         res = self.onConnect(request)
         #if yields(res):
         #  res = yield from res
      except http.HttpException as exc:
         self.failHandshake(exc.reason, exc.code)
      except Exception as exc:
         self.failHandshake(http.INTERNAL_SERVER_ERROR[1], http.INTERNAL_SERVER_ERROR[0])
      else:
         self.succeedHandshake(res)



class WebSocketClientProtocol(WebSocketAdapterProtocol, protocol.WebSocketClientProtocol):
   """
   Base class for Asyncio WebSocket client protocols.
   """

   def _onConnect(self, response):
      res = self.onConnect(response)
      if yields(res):
         asyncio.async(res)



class WebSocketAdapterFactory:
   """
   Adapter class for Asyncio WebSocket client and server factories.
   """

   def _log(self, msg):
      print(msg)


   def _callLater(self, delay, fun):
      return self.loop.call_later(delay, fun)


   def __call__(self):
      proto = self.protocol()
      proto.factory = self
      return proto



class WebSocketServerFactory(WebSocketAdapterFactory, protocol.WebSocketServerFactory):
   """
   Base class for Asyncio WebSocket server factories.
   """

   def __init__(self, *args, **kwargs):
      """
      In addition to all arguments to the constructor of
      :class:`autobahn.websocket.protocol.WebSocketServerFactory`,
      you can supply a `loop` keyword argument to specify the
      Asyncio loop to be used.
      """
      if 'loop' in kwargs:
         if kwargs['loop']:
            self.loop = kwargs['loop']
         else:
            self.loop = asyncio.get_event_loop()
         del kwargs['loop']
      else:
         self.loop = asyncio.get_event_loop()

      protocol.WebSocketServerFactory.__init__(self, *args, **kwargs)



class WebSocketClientFactory(WebSocketAdapterFactory, protocol.WebSocketClientFactory):
   """
   Base class for Asyncio WebSocket client factories.
   """

   def __init__(self, *args, **kwargs):
      """
      In addition to all arguments to the constructor of
      :class:`autobahn.websocket.protocol.WebSocketClientFactory`,
      you can supply a `loop` keyword argument to specify the
      Asyncio loop to be used.
      """
      if 'loop' in kwargs:
         if kwargs['loop']:
            self.loop = kwargs['loop']
         else:
            self.loop = asyncio.get_event_loop()
         del kwargs['loop']
      else:
         self.loop = asyncio.get_event_loop()

      protocol.WebSocketClientFactory.__init__(self, *args, **kwargs)



###############################################################################
##
##  Copyright (C) 2013 Tavendo GmbH
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##      http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.
##
###############################################################################

from __future__ import absolute_import

from zope.interface import implementer

from twisted.plugin import IPlugin
from twisted.internet.interfaces import IStreamServerEndpointStringParser, \
                                        IStreamServerEndpoint, \
                                        IStreamClientEndpointStringParser, \
                                        IStreamClientEndpoint

from twisted.internet.endpoints import serverFromString, clientFromString

from autobahn.twisted.websocket import WrappingWebSocketServerFactory, \
                                       WrappingWebSocketClientFactory


def _parseOptions(options):
   opts = {}

   if not 'url' in options:
      raise Exception("URL needed")
   else:
      opts['url'] = options['url']

   if 'compression' in options:
      value = options['compression'].lower().strip()
      if value == 'true':
         opts['enableCompression'] = True
      elif value == 'false':
         opts['enableCompression'] = False
      else:
         raise Exception("invalid value '{}' for compression".format(value))

   if 'autofrag' in options:
      try:
         value = int(options['autofrag'])
      except:
         raise Exception("invalid value '{}' for autofrag".format(options['autofrag']))

      if value < 0:
         raise Exception("negative value '{}' for autofrag".format(value))

      opts['autoFragmentSize'] = value

   if 'debug' in options:
      value = options['debug'].lower().strip()
      if value == 'true':
         opts['debug'] = True
      elif value == 'false':
         opts['debug'] = False
      else:
         raise Exception("invalid value '{}' for debug".format(value))

   return opts


@implementer(IPlugin)
@implementer(IStreamServerEndpointStringParser)
class AutobahnServerParser(object):

   prefix = "autobahn"

   def parseStreamServer(self, reactor, description, **options):

      ## The present endpoint plugin is intended to be used as in the
      ## following for running a streaming protocol over WebSocket over
      ## an underlying stream transport.
      ##
      ##  endpoint = serverFromString(reactor,
      ##    "autobahn:tcp\:9000\:interface\=0.0.0.0:url=ws\://localhost\:9000:compress=false"
      ##
      ## This will result in `parseStreamServer` to be called will
      ##
      ##   description == tcp:9000:interface=0.0.0.0
      ##
      ## and
      ##
      ##   options == {'url': 'ws://localhost:9000', 'compress': 'false'}
      ##
      ## Essentially, we are using the `\:` escape to coerce the endpoint descriptor
      ## of the underyling stream transport into one (first) positional argument.
      ##
      ## Note that the `\:` within "url" is another form of escaping!
      ##
      opts = _parseOptions(options)
      endpoint = serverFromString(reactor, description)
      return AutobahnServerEndpoint(reactor, endpoint, opts)


@implementer(IPlugin)
@implementer(IStreamServerEndpoint)
class AutobahnServerEndpoint(object):

   def __init__(self, reactor, endpoint, options):
      self._reactor = reactor
      self._endpoint = endpoint
      self._options = options

   def listen(self, protocolFactory):
      return self._endpoint.listen(WrappingWebSocketServerFactory(protocolFactory, reactor = self._reactor, **self._options))


@implementer(IPlugin)
@implementer(IStreamClientEndpointStringParser)
class AutobahnClientParser(object):

   prefix = "autobahn"

## <oberstet> Is there are particular reason why "plugin.parseStreamServer" provides a "reactor" argument while "plugin.parseStreamClient" does not?
## <oberstet> http://twistedmatrix.com/trac/browser/tags/releases/twisted-13.2.0/twisted/internet/endpoints.py#L1396
## <oberstet> http://twistedmatrix.com/trac/browser/tags/releases/twisted-13.2.0/twisted/internet/endpoints.py#L1735

## => http://twistedmatrix.com/trac/ticket/4956
## => https://twistedmatrix.com/trac/ticket/5069

#   def parseStreamClient(self, reactor, description, **options):
   def parseStreamClient(self, description, **options):
      from twisted.internet import reactor
      opts = _parseOptions(options)
      endpoint = clientFromString(reactor, description)
      return AutobahnClientEndpoint(reactor, endpoint, opts)


@implementer(IPlugin)
@implementer(IStreamClientEndpoint)
class AutobahnClientEndpoint(object):

   def __init__(self, reactor, endpoint, options):
      self._reactor = reactor
      self._endpoint = endpoint
      self._options = options

   def connect(self, protocolFactory):
      return self._endpoint.connect(WrappingWebSocketClientFactory(protocolFactory, reactor = self._reactor, **self._options))


autobahnServerParser = AutobahnServerParser()
autobahnClientParser = AutobahnClientParser()



###############################################################################
##
##  Copyright (C) 2013 Tavendo GmbH
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##      http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.
##
###############################################################################


## incoming client connections
## incoming broker connections
## outgoing broker connections

from autobahn.util import newid

from message import WampMessageSubscribe, \
                    WampMessageSubscription, \
                    WampMessageSubscribeError, \
                    WampMessageUnsubscribe, \
                    WampMessagePublish, \
                    WampMessageEvent, \
                    WampMessageMetaEvent


class Broker:

   def __init__(self):
      ## FIXME: maintain 2 maps: topic => protos (subscribers), proto => topics
      self._protos = set()
      self._subscribers = {}
      self._brokers = set()

   def addBroker(self, proto):
      assert(proto not in self._brokers)
      self._brokers.add(proto)

      # def on_publish(topic, event):

      # for topic in self._subscribers:
      #    proto.subscribe()
      #    msg = WampMessageEvent(publish.topic, publish.event)
      #    bytes, isbinary = proto.factory._serializer.serialize(msg)
      #    for proto in receivers:
      #       proto.sendMessage(bytes, isbinary)


   def removeBroker(self, proto):
      assert(proto in self._brokers)
      self._brokers.remove(proto)

   def add(self, proto):
      assert(proto not in self._protos)
      self._protos.add(proto)

   def remove(self, proto):
      assert(proto in self._protos)
      self._protos.remove(proto)
      for subscriptionid, subscribers in self._subscribers.values():
         subscribers.discard(proto)


   def onPublish(self, proto, publish):
      assert(proto in self._protos)

      for broker_proto in self._brokers:
         if broker_proto != proto:
            bytes, isbinary = broker_proto._serializer.serialize(publish)
            broker_proto.sendMessage(bytes, isbinary)

      if self._subscribers.has_key(publish.topic):
         subscriptionid, receivers = self._subscribers[publish.topic]
         if len(receivers) > 0:
            msg = WampMessageEvent(subscriptionid, publish.topic, publish.event)
            for proto in receivers:
               bytes, isbinary = proto._serializer.serialize(msg)
               proto.sendMessage(bytes, isbinary)


   def onSubscribe(self, proto, subscribe):
      assert(proto in self._protos)

      # if subscribe.topic.startswith("http://example.com/"):
      #    proto.sendWampMessage(WampMessageSubscribeError(subscribe.subscribeid, "http://api.wamp.ws/error#forbidden"))
      #    return


      if not self._subscribers.has_key(subscribe.topic):
         subscriptionid = newid()
         self._subscribers[subscribe.topic] = (subscriptionid, set())

      subscriptionid, subscribers = self._subscribers[subscribe.topic]

      if not proto in subscribers:
         subscribers.add(proto)

      proto.sendWampMessage(WampMessageSubscription(subscribe.subscribeid, subscriptionid))


   def onPublish2(self, proto, publish):
      assert(proto in self._protos)

      for broker_proto in self._brokers:
         if broker_proto != proto:
            bytes, isbinary = broker_proto._serializer.serialize(publish)
            broker_proto.sendMessage(bytes, isbinary)

      if self._subscribers.has_key(publish.topic):
         receivers = self._subscribers[publish.topic]
         if len(receivers) > 0:
            msg = WampMessageEvent(publish.topic, publish.event)
            bytes, isbinary = proto._serializer.serialize(msg)
            for proto in receivers:
               proto.sendMessage(bytes, isbinary)


   def onSubscribe2(self, proto, subscribe):
      assert(proto in self._protos)

      if not self._subscribers.has_key(subscribe.topic):
         self._subscribers[subscribe.topic] = set()

      if not proto in self._subscribers[subscribe.topic]:
         self._subscribers[subscribe.topic].add(proto)


   def onUnsubscribe(self, proto, unsubscribe):
      assert(proto in self._protos)

      if self._subscribers.has_key(unsubscribe.topic):
         if proto in self._subscribers[unsubscribe.topic]:
            self._subscribers[unsubscribe.topic].discard(proto)




###############################################################################
##
##  Copyright (C) 2011-2013 Tavendo GmbH
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##      http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.
##
###############################################################################

from __future__ import absolute_import

from setuptools import setup
from distutils import log


import sys
PY3 = sys.version_info.major > 2
if PY3:
   log.info("Python 3 detected.")
else:
   log.info("Python 2 detected.")


LONGSDESC = """
Autobahn|Python provides implementations of

 * The WebSocket Protocol
 * The Web Application Messaging Protocol (WAMP)

for Twisted and Asyncio.

More information:

 * https://github.com/tavendo/AutobahnPython
 * http://autobahn.ws/python
 * http://wamp.ws
"""

## get version string from "autobahn/__init__.py"
## See: http://stackoverflow.com/a/7071358/884770
##
import re
VERSIONFILE="autobahn/__init__.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
   verstr = mo.group(1)
else:
   raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))


## Autobahn core packages
##
packages = ['autobahn',
            'autobahn.wamp2',
            'autobahn.websocket']


## Twisted integration
##
try:
   from twisted.internet import reactor
except:
   log.warn("Twisted not installed - skipping Twisted support packages")
   HAS_TWISTED = False
else:
   log.info("Twisted detected - installing Autobahn/Twisted integration")
   packages.append('autobahn.twisted')
   packages.append('twisted.plugins')
   HAS_TWISTED = True


## Asyncio integration
##
if PY3:
   try:
      from asyncio import Future
   except:
      log.warn("Asyncio not available - skipping Asyncio support packages")
   else:
      log.info("Asyncio detected - installing Autobahn/Asyncio integration")
      packages.append('autobahn.asyncio')
else:
   log.warn("Python 2 detected - skipping Autobahn/Asyncio integration")


## Now install Autobahn ..
##
setup(
   name = 'autobahn',
   version = verstr,
   description = 'Autobahn|Python provides WebSocket and WAMP for Twisted and Asyncio',
   long_description = LONGSDESC,
   license = 'Apache License 2.0',
   author = 'Tavendo GmbH',
   author_email = 'autobahnws@googlegroups.com',
   url = 'http://autobahn.ws/python',
   platforms = ('Any'),
   install_requires = ['setuptools', 'zope.interface>=4.0.2'],
   extras_require = {
      ## this is ONLY needed on Python 3.3. Python 3.4+ has that integrated already!
      'asyncio': ["asyncio>=0.2.1"],

      ## you need Twisted for Autobahn/Twisted - obviously
      'twisted': ["twisted>=11.1"],

      ## native WebSocket and JSON acceleration: this should ONLY be used on CPython
      'accelerate': ["wsaccel>=0.6.2", "ujson>=1.33"],

      ## for (non-standard) WebSocket compression methods - not needed if you
      ## only want standard WebSocket compression ("permessage-deflate")
      'compress': ["python-snappy>=0.5", "lz4>=0.2.1"],

      ## needed if you want WAMPv2 binary serialization support
      'binary': ["msgpack-python>=0.4.0"]
   },
   packages = packages,
   zip_safe = False,
   ## http://pypi.python.org/pypi?%3Aaction=list_classifiers
   ##
   classifiers = ["License :: OSI Approved :: Apache Software License",
                  "Development Status :: 5 - Production/Stable",
                  "Environment :: Console",
                  "Framework :: Twisted",
                  "Intended Audience :: Developers",
                  "Operating System :: OS Independent",
                  "Programming Language :: Python",
                  "Topic :: Internet",
                  "Topic :: Software Development :: Libraries"],
   keywords = 'autobahn autobahn.ws websocket realtime rfc6455 wamp rpc pubsub twisted asyncio'
)


if HAS_TWISTED:
   # Make Twisted regenerate the dropin.cache, if possible. This is necessary
   # because in a site-wide install, dropin.cache cannot be rewritten by
   # normal users.
   try:
      from twisted.plugin import IPlugin, getPlugins
      list(getPlugins(IPlugin))
   except Exception as e:
      log.warn("Failed to update Twisted plugin cache: {}".format(e))
   else:
      log.info("Twisted dropin.cache regenerated.")

   ## verify that Autobahn Twisted endpoints have been installed
   try:
      from twisted.internet.interfaces import IStreamServerEndpointStringParser
      from twisted.internet.interfaces import IStreamClientEndpointStringParser

      has_server_endpoint = False
      for plugin in getPlugins(IStreamServerEndpointStringParser):
         if plugin.prefix == "autobahn":
            has_server_endpoint = True
            break

      if has_server_endpoint:
         log.info("Autobahn Twisted stream server endpoint successfully installed")
      else:
         log.warn("Autobahn Twisted stream server endpoint installation seems to have failed")

      has_client_endpoint = False
      for plugin in getPlugins(IStreamClientEndpointStringParser):
         if plugin.prefix == "autobahn":
            has_client_endpoint = True
            break

      if has_client_endpoint:
         log.info("Autobahn Twisted stream client endpoint successfully installed")
      else:
         log.warn("Autobahn Twisted stream client endpoint installation seems to have failed")

   except:
      log.warn("Autobahn Twisted endpoint installation could not be verified")



###############################################################################
##
##  Copyright (C) 2011-2013 Tavendo GmbH
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##      http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.
##
###############################################################################

from __future__ import absolute_import

__all__ = ['WebSocketServerProtocol',
           'WebSocketServerFactory',
           'WebSocketClientProtocol',
           'WebSocketClientFactory',
           'WrappingWebSocketServerFactory',
           'WrappingWebSocketClientFactory',
           'listenWS',
           'connectWS']

from base64 import b64encode, b64decode

from zope.interface import implementer

import twisted.internet.protocol
from twisted.internet.defer import maybeDeferred
from twisted.python import log
from twisted.internet.interfaces import ITransport

from autobahn.websocket import protocol
from autobahn.websocket import http

from autobahn.websocket.compress import PerMessageDeflateOffer, \
                                        PerMessageDeflateOfferAccept, \
                                        PerMessageDeflateResponse, \
                                        PerMessageDeflateResponseAccept


class WebSocketAdapterProtocol(twisted.internet.protocol.Protocol):
   """
   Adapter class for Twisted WebSocket client and server protocols.
   """

   def connectionMade(self):
      ## the peer we are connected to
      peer = self.transport.getPeer()
      try:
         self.peer = "%s:%d" % (peer.host, peer.port)
      except:
         ## eg Unix Domain sockets don't have host/port
         self.peer = str(peer)

      self._connectionMade()

      ## Set "Nagle"
      try:
         self.transport.setTcpNoDelay(self.tcpNoDelay)
      except:
         ## eg Unix Domain sockets throw Errno 22 on this
         pass


   def connectionLost(self, reason):
      self._connectionLost(reason)


   def dataReceived(self, data):
      self._dataReceived(data)


   def _closeConnection(self, abort = False):
      if abort:
         self.transport.abortConnection()
      else:
         self.transport.loseConnection()


   def _onOpen(self):
      self.onOpen()

   def _onMessageBegin(self, isBinary):
      self.onMessageBegin(isBinary)

   def _onMessageFrameBegin(self, length):
      self.onMessageFrameBegin(length)

   def _onMessageFrameData(self, payload):
      self.onMessageFrameData(payload)

   def _onMessageFrameEnd(self):
      self.onMessageFrameEnd()

   def _onMessageFrame(self, payload):
      self.onMessageFrame(payload)

   def _onMessageEnd(self):
      self.onMessageEnd()

   def _onMessage(self, payload, isBinary):
      self.onMessage(payload, isBinary)

   def _onPing(self, payload):
      self.onPing(payload)

   def _onPong(self, payload):
      self.onPong(payload)

   def _onClose(self, wasClean, code, reason):
      self.onClose(wasClean, code, reason)


   def registerProducer(self, producer, streaming):
      """
      Register a Twisted producer with this protocol.

      Modes: Hybi, Hixie

      :param producer: A Twisted push or pull producer.
      :type producer: object
      :param streaming: Producer type.
      :type streaming: bool
      """
      self.transport.registerProducer(producer, streaming)



class WebSocketServerProtocol(WebSocketAdapterProtocol, protocol.WebSocketServerProtocol):
   """
   Base class for Twisted WebSocket server protocols.
   """

   def _onConnect(self, request):
      ## onConnect() will return the selected subprotocol or None
      ## or a pair (protocol, headers) or raise an HttpException
      ##
      res = maybeDeferred(self.onConnect, request)

      res.addCallback(self.succeedHandshake)

      def forwardError(failure):
         if failure.check(http.HttpException):
            return self.failHandshake(failure.value.reason, failure.value.code)
         else:
            if self.debug:
               self.factory._log("Unexpected exception in onConnect ['%s']" % failure.value)
            return self.failHandshake(http.INTERNAL_SERVER_ERROR[1], http.INTERNAL_SERVER_ERROR[0])

      res.addErrback(forwardError)



class WebSocketClientProtocol(WebSocketAdapterProtocol, protocol.WebSocketClientProtocol):
   """
   Base class for Twisted WebSocket client protocols.
   """

   def _onConnect(self, response):
      self.onConnect(response)



class WebSocketAdapterFactory:
   """
   Adapter class for Twisted WebSocket client and server factories.
   """

   def _log(self, msg):
      log.msg(msg)


   def _callLater(self, delay, fun):
      return self.reactor.callLater(delay, fun)



class WebSocketServerFactory(WebSocketAdapterFactory, protocol.WebSocketServerFactory, twisted.internet.protocol.ServerFactory):
   """
   Base class for Twisted WebSocket server factories.

   .. seealso:: `twisted.internet.protocol.ServerFactory <http://twistedmatrix.com/documents/current/api/twisted.internet.protocol.ServerFactory.html>`_
   """

   def __init__(self, *args, **kwargs):
      """
      In addition to all arguments to the constructor of
      :class:`autobahn.websocket.protocol.WebSocketServerFactory`,
      you can supply a `reactor` keyword argument to specify the 
      Twisted reactor to be used.
      """
      ## lazy import to avoid reactor install upon module import
      if 'reactor' in kwargs:
         if kwargs['reactor']:
            self.reactor = kwargs['reactor']
         else:
            from twisted.internet import reactor
            self.reactor = reactor
         del kwargs['reactor']
      else:
         from twisted.internet import reactor
         self.reactor = reactor

      protocol.WebSocketServerFactory.__init__(self, *args, **kwargs)



class WebSocketClientFactory(WebSocketAdapterFactory, protocol.WebSocketClientFactory, twisted.internet.protocol.ClientFactory):
   """
   Base class for Twisted WebSocket client factories.

   .. seealso:: `twisted.internet.protocol.ClientFactory <http://twistedmatrix.com/documents/current/api/twisted.internet.protocol.ClientFactory.html>`_
   """

   def __init__(self, *args, **kwargs):
      """
      In addition to all arguments to the constructor of
      :class:`autobahn.websocket.protocol.WebSocketClientFactory`,
      you can supply a `reactor` keyword argument to specify the 
      Twisted reactor to be used.
      """
      ## lazy import to avoid reactor install upon module import
      if 'reactor' in kwargs:
         if kwargs['reactor']:
            self.reactor = kwargs['reactor']
         else:
            from twisted.internet import reactor
            self.reactor = reactor
         del kwargs['reactor']
      else:
         from twisted.internet import reactor
         self.reactor = reactor

      protocol.WebSocketClientFactory.__init__(self, *args, **kwargs)



@implementer(ITransport)
class WrappingWebSocketAdapter:
   """
   An adapter for stream-based transport over WebSocket.

   This follows "websockify" (https://github.com/kanaka/websockify)
   and should be compatible with that.

   It uses WebSocket subprotocol negotiation and 2 subprotocols:
     - binary
     - base64

   Octets are either transmitted as the payload of WebSocket binary
   messages when using the 'binary' subprotocol, or encoded with Base64
   and then transmitted as the payload of WebSocket text messages when
   using the 'base64' subprotocol.
   """

   def onConnect(self, requestOrResponse):

      ## Negotiate either the 'binary' or the 'base64' WebSocket subprotocol
      ##
      if isinstance(requestOrResponse, protocol.ConnectionRequest):
         request = requestOrResponse
         for p in request.protocols:
            if p in ['binary', 'base64']:
               self._binaryMode = (p == 'binary')
               return p
         raise http.HttpException(http.NOT_ACCEPTABLE[0], "this server only speaks 'binary' and 'base64' WebSocket subprotocols")
      elif isinstance(requestOrResponse, protocol.ConnectionResponse):
         response = requestOrResponse
         if response.protocol not in ['binary', 'base64']:
            self.failConnection(protocol.WebSocketProtocol.CLOSE_STATUS_CODE_PROTOCOL_ERROR, "this client only speaks 'binary' and 'base64' WebSocket subprotocols")
         self._binaryMode = (response.protocol == 'binary')
      else:
         ## should not arrive here
         raise Exception("logic error")

   def onOpen(self):
      self._proto.connectionMade()

   def onMessage(self, payload, isBinary):
      if isBinary != self._binaryMode:
         self.failConnection(protocol.WebSocketProtocol.CLOSE_STATUS_CODE_UNSUPPORTED_DATA, "message payload type does not match the negotiated subprotocol")
      else:
         if not isBinary:
            try:
               payload = b64decode(payload)
            except Exception as e:
               self.failConnection(protocol.WebSocketProtocol.CLOSE_STATUS_CODE_INVALID_PAYLOAD, "message payload base64 decoding error: {}".format(e))
         #print("forwarding payload: {}".format(binascii.hexlify(payload)))
         self._proto.dataReceived(payload)

   def onClose(self, wasClean, code, reason):
      self._proto.connectionLost(None)

   def write(self, data):
      #print("sending payload: {}".format(binascii.hexlify(data)))
      ## part of ITransport
      assert(type(data) == bytes)
      if self._binaryMode:
         self.sendMessage(data, isBinary = True)
      else:
         data = b64encode(data)
         self.sendMessage(data, isBinary = False)

   def writeSequence(self, data):
      ## part of ITransport
      for d in data:
         self.write(data)

   def loseConnection(self):
      ## part of ITransport
      self.sendClose()



class WrappingWebSocketServerProtocol(WrappingWebSocketAdapter, WebSocketServerProtocol):
   """
   Server protocol for stream-based transport over WebSocket.
   """



class WrappingWebSocketClientProtocol(WrappingWebSocketAdapter, WebSocketClientProtocol):
   """
   Client protocol for stream-based transport over WebSocket.
   """



class WrappingWebSocketServerFactory(WebSocketServerFactory):
   """
   Wrapping server factory for stream-based transport over WebSocket.
   """

   def __init__(self,
                factory,
                url,
                reactor = None,
                enableCompression = True,
                autoFragmentSize = 0,
                debug = False):
      """
      Constructor.

      :param factory: Stream-based factory to be wrapped.
      :type factory: A subclass of `twisted.internet.protocol.Factory`
      :param url: WebSocket URL of the server this server factory will work for.
      :type url: str
      """
      self._factory = factory

      WebSocketServerFactory.__init__(self,
         url = url,
         reactor = reactor,
         protocols = ['binary', 'base64'],
         debug = debug)

      ## automatically fragment outgoing traffic into WebSocket frames
      ## of this size
      self.setProtocolOptions(autoFragmentSize = autoFragmentSize)

      ## play nice and perform WS closing handshake
      self.setProtocolOptions(failByDrop = False)

      if enableCompression:
         ## Enable WebSocket extension "permessage-deflate".
         ##

         ## Function to accept offers from the client ..
         def accept(offers):
            for offer in offers:
               if isinstance(offer, PerMessageDeflateOffer):
                  return PerMessageDeflateOfferAccept(offer)

         self.setProtocolOptions(perMessageCompressionAccept = accept)


   def buildProtocol(self, addr):
      proto = WrappingWebSocketServerProtocol()
      proto.factory = self
      proto._proto = self._factory.buildProtocol(addr)
      proto._proto.transport = proto
      return proto


   def startFactory(self):
      self._factory.startFactory()
      WebSocketServerFactory.startFactory(self)


   def stopFactory(self):
      self._factory.stopFactory()
      WebSocketServerFactory.stopFactory(self)



class WrappingWebSocketClientFactory(WebSocketClientFactory):
   """
   Wrapping client factory for stream-based transport over WebSocket.
   """

   def __init__(self,
                factory,
                url,
                reactor = None,
                enableCompression = True,
                autoFragmentSize = 0,
                debug = False):
      """
      Constructor.

      :param factory: Stream-based factory to be wrapped.
      :type factory: A subclass of `twisted.internet.protocol.Factory`
      :param url: WebSocket URL of the server this client factory will connect to.
      :type url: str
      """
      self._factory = factory

      WebSocketClientFactory.__init__(self,
         url = url,
         reactor = reactor,
         protocols = ['binary', 'base64'],
         debug = debug)

      ## automatically fragment outgoing traffic into WebSocket frames
      ## of this size
      self.setProtocolOptions(autoFragmentSize = autoFragmentSize)

      ## play nice and perform WS closing handshake
      self.setProtocolOptions(failByDrop = False)

      if enableCompression:
         ## Enable WebSocket extension "permessage-deflate".
         ##

         ## The extensions offered to the server ..
         offers = [PerMessageDeflateOffer()]
         self.setProtocolOptions(perMessageCompressionOffers = offers)

         ## Function to accept responses from the server ..
         def accept(response):
            if isinstance(response, PerMessageDeflateResponse):
               return PerMessageDeflateResponseAccept(response)

         self.setProtocolOptions(perMessageCompressionAccept = accept)


   def buildProtocol(self, addr):
      proto = WrappingWebSocketClientProtocol()
      proto.factory = self
      proto._proto = self._factory.buildProtocol(addr)
      proto._proto.transport = proto
      return proto



def connectWS(factory, contextFactory = None, timeout = 30, bindAddress = None):
   """
   Establish WebSocket connection to a server. The connection parameters like target
   host, port, resource and others are provided via the factory.

   :param factory: The WebSocket protocol factory to be used for creating client protocol instances.
   :type factory: An :class:`autobahn.websocket.WebSocketClientFactory` instance.
   :param contextFactory: SSL context factory, required for secure WebSocket connections ("wss").
   :type contextFactory: A `twisted.internet.ssl.ClientContextFactory <http://twistedmatrix.com/documents/current/api/twisted.internet.ssl.ClientContextFactory.html>`_ instance.
   :param timeout: Number of seconds to wait before assuming the connection has failed.
   :type timeout: int
   :param bindAddress: A (host, port) tuple of local address to bind to, or None.
   :type bindAddress: tuple

   :returns: obj -- An object which implements `twisted.interface.IConnector <http://twistedmatrix.com/documents/current/api/twisted.internet.interfaces.IConnector.html>`_.
   """
   ## lazy import to avoid reactor install upon module import
   if hasattr(factory, 'reactor'):
      reactor = factory.reactor
   else:
      from twisted.internet import reactor

   if factory.proxy is not None:
      if factory.isSecure:
         raise Exception("WSS over explicit proxies not implemented")
      else:
         conn = reactor.connectTCP(factory.proxy['host'], factory.proxy['port'], factory, timeout, bindAddress)
   else:
      if factory.isSecure:
         if contextFactory is None:
            # create default client SSL context factory when none given
            from twisted.internet import ssl
            contextFactory = ssl.ClientContextFactory()
         conn = reactor.connectSSL(factory.host, factory.port, factory, contextFactory, timeout, bindAddress)
      else:
         conn = reactor.connectTCP(factory.host, factory.port, factory, timeout, bindAddress)
   return conn



def listenWS(factory, contextFactory = None, backlog = 50, interface = ''):
   """
   Listen for incoming WebSocket connections from clients. The connection parameters like
   listening port and others are provided via the factory.

   :param factory: The WebSocket protocol factory to be used for creating server protocol instances.
   :type factory: An :class:`autobahn.websocket.WebSocketServerFactory` instance.
   :param contextFactory: SSL context factory, required for secure WebSocket connections ("wss").
   :type contextFactory: A twisted.internet.ssl.ContextFactory.
   :param backlog: Size of the listen queue.
   :type backlog: int
   :param interface: The interface (derived from hostname given) to bind to, defaults to '' (all).
   :type interface: str

   :returns: obj -- An object that implements `twisted.interface.IListeningPort <http://twistedmatrix.com/documents/current/api/twisted.internet.interfaces.IListeningPort.html>`_.
   """
   ## lazy import to avoid reactor install upon module import
   if hasattr(factory, 'reactor'):
      reactor = factory.reactor
   else:
      from twisted.internet import reactor

   if factory.isSecure:
      if contextFactory is None:
         raise Exception("Secure WebSocket listen requested, but no SSL context factory given")
      listener = reactor.listenSSL(factory.port, factory, contextFactory, backlog, interface)
   else:
      listener = reactor.listenTCP(factory.port, factory, backlog, interface)
   return listener



###############################################################################
##
##  Copyright (C) 2013 Tavendo GmbH
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##      http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.
##
###############################################################################

__all__= ['WampWebSocketServerProtocol',
          'WampWebSocketClientProtocol',
          'WampWebSocketServerFactory',
          'WampWebSocketClientFactory']


from autobahn.websocket import WebSocketServerProtocol, \
                               WebSocketServerFactory, \
                               WebSocketClientProtocol, \
                               WebSocketClientFactory

from autobahn.websocket import HttpException
from autobahn.httpstatus import HTTP_STATUS_CODE_BAD_REQUEST

from protocol import WampProtocol, parseSubprotocolIdentifier
from serializer import WampJsonSerializer, WampMsgPackSerializer



class WampWebSocketServerProtocol(WampProtocol, WebSocketServerProtocol):

   def onConnect(self, connectionRequest):
      headers = {}
      for subprotocol in connectionRequest.protocols:
         version, serializerId = parseSubprotocolIdentifier(subprotocol)
         if version == 2 and serializerId in self.factory._serializers.keys():
            self._serializer = self.factory._serializers[serializerId]
            return subprotocol, headers

      raise HttpException(HTTP_STATUS_CODE_BAD_REQUEST[0], "This server only speaks WebSocket subprotocols %s" % ', '.join(self.factory.protocols))




class WampWebSocketClientProtocol(WampProtocol, WebSocketClientProtocol):

   def onConnect(self, connectionResponse):
      if connectionResponse.protocol not in self.factory.protocols:
         raise Exception("Server does not speak any of the WebSocket subprotocols we requested (%s)." % ', '.join(self.factory.protocols))

      version, serializerId = parseSubprotocolIdentifier(connectionResponse.protocol)
      self._serializer = self.factory._serializers[serializerId]



class WampWebSocketServerFactory(WebSocketServerFactory):

   protocol = WampWebSocketServerProtocol

   def __init__(self,
                url,
                debug = False,
                serializers = None,
                reactor = None):

      if serializers is None:
         serializers = [WampMsgPackSerializer(), WampJsonSerializer()]

      self._serializers = {}
      for ser in serializers:
         self._serializers[ser.SERIALIZER_ID] = ser

      protocols = ["wamp.2.%s" % ser.SERIALIZER_ID for ser in serializers]

      WebSocketServerFactory.__init__(self,
                                      url,
                                      debug = debug,
                                      protocols = protocols,
                                      reactor = reactor)



class WampWebSocketClientFactory(WebSocketClientFactory):

   protocol = WampWebSocketClientProtocol

   def __init__(self,
                url,
                debug = False,
                serializers = None,
                reactor = None):

      if serializers is None:
         serializers = [WampMsgPackSerializer(), WampJsonSerializer()]

      self._serializers = {}
      for ser in serializers:
         self._serializers[ser.SERIALIZER_ID] = ser

      protocols = ["wamp.2.%s" % ser.SERIALIZER_ID for ser in serializers]

      WebSocketClientFactory.__init__(self,
                                      url,
                                      debug = debug,
                                      protocols = protocols,
                                      reactor = reactor)



###############################################################################
##
##  Copyright (C) 2013 Tavendo GmbH
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##      http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.
##
###############################################################################

__all__ = ['WampSerializer',
           'JsonSerializer',
           'WampJsonSerializer']


from zope.interface import implementer

from interfaces import ISerializer
from error import WampProtocolError
from message import WampMessageHello, \
                    WampMessageHeartbeat, \
                    WampMessageRoleChange, \
                    WampMessageSubscribe, \
                    WampMessageSubscription, \
                    WampMessageSubscribeError, \
                    WampMessageUnsubscribe, \
                    WampMessagePublish, \
                    WampMessageEvent, \
                    WampMessageMetaEvent, \
                    WampMessageProvide, \
                    WampMessageUnprovide, \
                    WampMessageCall, \
                    WampMessageCancelCall, \
                    WampMessageCallProgress, \
                    WampMessageCallResult, \
                    WampMessageCallError



class WampSerializer:
   """
   WAMP serializer is the core glue between parsed WAMP message objects and the
   bytes on wire (the transport).
   """

   MESSAGE_TYPE_MAP = {
      ## Session
      WampMessageHello.MESSAGE_TYPE:            WampMessageHello,
      WampMessageHeartbeat.MESSAGE_TYPE:        WampMessageHeartbeat,
      WampMessageRoleChange.MESSAGE_TYPE:       WampMessageRoleChange,

      ## PubSub
      WampMessageSubscribe.MESSAGE_TYPE:        WampMessageSubscribe,
      WampMessageSubscription.MESSAGE_TYPE:     WampMessageSubscription,
      WampMessageSubscribeError.MESSAGE_TYPE:   WampMessageSubscribeError,

      WampMessageUnsubscribe.MESSAGE_TYPE:      WampMessageUnsubscribe,
      WampMessagePublish.MESSAGE_TYPE:          WampMessagePublish,
      WampMessageEvent.MESSAGE_TYPE:            WampMessageEvent,
      WampMessageMetaEvent.MESSAGE_TYPE:        WampMessageMetaEvent,

      ## RPC
      WampMessageProvide.MESSAGE_TYPE:          WampMessageProvide,
      WampMessageUnprovide.MESSAGE_TYPE:        WampMessageUnprovide,
      WampMessageCall.MESSAGE_TYPE:             WampMessageCall,
      WampMessageCancelCall.MESSAGE_TYPE:       WampMessageCancelCall,
      WampMessageCallProgress.MESSAGE_TYPE:     WampMessageCallProgress,
      WampMessageCallResult.MESSAGE_TYPE:       WampMessageCallResult,
      WampMessageCallError.MESSAGE_TYPE:        WampMessageCallError,
   }


   def __init__(self, serializer):
      """
      Constructor.

      :param serializer: The wire serializer to use for WAMP wire processing.
      :type serializer: An object that implements :class:`autobahn.interfaces.ISerializer`.
      """
      self._serializer = serializer


   def serialize(self, wampMessage):
      """
      Serializes a WAMP message to bytes to be sent to a transport.

      :param wampMessage: An instance of a subclass of :class:`autobahn.wamp2message.WampMessage`.
      :type wampMessage: obj
      :returns str -- A byte string.
      """
      return wampMessage.serialize(self._serializer), self._serializer.isBinary


   def unserialize(self, bytes, isBinary):
      """
      Unserializes bytes from a transport and parses a WAMP message.

      :param bytes: Byte string from wire.
      :type bytes: str or bytes
      :returns obj -- An instance of a subclass of :class:`autobahn.wamp2message.WampMessage`.
      """
      if isBinary != self._serializer.isBinary:
         raise WampProtocolError("invalid serialization of WAMP message [binary = %s, but expected %s]" % (isBinary, self._serializer.isBinary))

      try:
         raw_msg = self._serializer.unserialize(bytes)
      except Exception as e:
         raise WampProtocolError("invalid serialization of WAMP message [%s]" % e)

      if type(raw_msg) != list:
         raise WampProtocolError("invalid type %s for WAMP message" % type(raw_msg))

      if len(raw_msg) == 0:
         raise WampProtocolError("missing message type in WAMP message")

      message_type = raw_msg[0]

      if type(message_type) != int:
         raise WampProtocolError("invalid type %d for WAMP message type" % type(message_type))

      Klass = self.MESSAGE_TYPE_MAP.get(message_type)

      if Klass is None:
         raise WampProtocolError("invalid WAMP message type %d" % message_type)

      msg = Klass.parse(raw_msg)

      return msg



import json

@implementer(ISerializer)
class JsonSerializer:

   isBinary = False

   def serialize(self, obj):
      return json.dumps(obj, separators = (',',':'))


   def unserialize(self, bytes):
      return json.loads(bytes)



class WampJsonSerializer(WampSerializer):

   SERIALIZER_ID = "json"

   def __init__(self):
      WampSerializer.__init__(self, JsonSerializer())



try:
   import msgpack
except:
   pass
else:
   @implementer(ISerializer)
   class MsgPackSerializer:

      isBinary = True

      def serialize(self, obj):
         return msgpack.packb(obj, use_bin_type = True)


      def unserialize(self, bytes):
         return msgpack.unpackb(bytes, encoding = 'utf-8')

   __all__.append('MsgPackSerializer')


   class WampMsgPackSerializer(WampSerializer):

      SERIALIZER_ID = "msgpack"

      def __init__(self):
         WampSerializer.__init__(self, MsgPackSerializer())

   __all__.append('WampMsgPackSerializer')
